@@ -2,6 +2,7 @@
  *
  * Copyright (c) 2000-2003 Intel Corporation 
  * All rights reserved. 
+ * Copyright (c) 2012 France Telecom All rights reserved. 
  *
  * Redistribution and use in source and binary forms, with or without 
  * modification, are permitted provided that the following conditions are met: 
@@ -43,6 +44,9 @@
 		#include <lwres/netdb.h>
 	#endif
 #endif
+#ifdef WIN32
+	#define snprintf _snprintf
+#endif
 #include <assert.h>
 
 
@@ -65,7 +69,7 @@ static int is_reserved(
 	/*! [in] Char to be matched for RESERVED characters. */
 	char in)
 {
-	if (strchr(RESERVED, in)) {
+	if (strchr(RESERVED, (int)in)) {
 		return 1;
 	} else {
 		return 0;
@@ -83,7 +87,7 @@ int is_mark(
 	/*! [in] Char to be matched for MARKED characters. */
 	char in)
 {
-	if (strchr(MARK, in)) {
+	if (strchr(MARK, (int)in)) {
 		return 1;
 	} else {
 		return 0;
@@ -132,17 +136,17 @@ int replace_escaped(char *in, size_t index, size_t *max)
 {
 	int tempInt = 0;
 	char tempChar = 0;
-	size_t i = 0;
-	size_t j = 0;
+	size_t i = (size_t)0;
+	size_t j = (size_t)0;
 
-	if (in[index] == '%' && isxdigit(in[index + 1]) && isxdigit(in[index + 2])) {
+	if (in[index] == '%' && isxdigit(in[index + (size_t)1]) && isxdigit(in[index + (size_t)2])) {
 		/* Note the "%2x", makes sure that we convert a maximum of two
 		 * characters. */
-		if (sscanf(&in[index + 1], "%2x", &tempInt) != 1) {
+		if (sscanf(&in[index + (size_t)1], "%2x", &tempInt) != 1) {
 			return 0;
 		}
 		tempChar = (char)tempInt;
-		for (i = index + 3, j = index; j < *max; i++, j++) {
+		for (i = index + (size_t)3, j = index; j < *max; i++, j++) {
 			in[j] = tempChar;
 			if (i < *max) {
 				tempChar = in[i];
@@ -150,7 +154,7 @@ int replace_escaped(char *in, size_t index, size_t *max)
 				tempChar = 0;
 			}
 		}
-		*max -= 2;
+		*max -= (size_t)2;
 		return 1;
 	} else {
 		return 0;
@@ -172,12 +176,12 @@ static size_t parse_uric(
 	/*! [out] Token object where the string of characters is copied. */
 	token *out)
 {
-	size_t i = 0;
+	size_t i = (size_t)0;
 
 	while (i < max &&
 	       (is_unreserved(in[i]) ||
 	        is_reserved(in[i])   ||
-	        ((i + 2 < max) && is_escaped(&in[i])))) {
+	        ((i + (size_t)2 < max) && is_escaped(&in[i])))) {
 		i++;
 	}
 
@@ -209,12 +213,12 @@ static void copy_token(
 
 int copy_URL_list(URL_list *in, URL_list *out)
 {
-    size_t len = strlen(in->URLs) + 1;
-    size_t i = 0;
+    size_t len = strlen(in->URLs) + (size_t)1;
+    size_t i = (size_t)0;
 
     out->URLs = NULL;
     out->parsedURLs = NULL;
-    out->size = 0;
+    out->size = (size_t)0;
 
     out->URLs = malloc(len);
     out->parsedURLs = malloc(sizeof(uri_type) * in->size);
@@ -222,7 +226,7 @@ int copy_URL_list(URL_list *in, URL_list *out)
     if ( !out->URLs || !out->parsedURLs)
         return UPNP_E_OUTOF_MEMORY;
     memcpy(out->URLs, in->URLs, len);
-    for( i = 0; i < in->size; i++ ) {
+    for( i = (size_t)0; i < in->size; i++ ) {
         /*copy the parsed uri */
         out->parsedURLs[i].type = in->parsedURLs[i].type;
         copy_token( &in->parsedURLs[i].scheme, in->URLs,
@@ -253,7 +257,7 @@ void free_URL_list(URL_list *list)
 	if (list->parsedURLs) {
 		free(list->parsedURLs);
 	}
-	list->size = 0;
+	list->size = (size_t)0;
 }
 
 
@@ -309,9 +313,18 @@ int token_cmp(token *in1, token *in2)
 		return memcmp(in1->buff, in2->buff, in1->size);
 }
 
-int parse_hostport(
+/*!
+ * \brief Parses a string representing a host and port (e.g. "127.127.0.1:80"
+ * or "localhost") and fills out a hostport_type struct with internet address
+ * and a token representing the full host and port.
+ *
+ * Uses gethostbyname.
+ */
+static int parse_hostport(
+	/*! [in] String of characters representing host and port. */
 	const char *in,
-	size_t max,
+	/*! [out] Output parameter where the host and port are represented as
+	 * an internet address. */
 	hostport_type *out)
 {
 	char workbuf[256];
@@ -328,8 +341,9 @@ int parse_hostport(
 	int ret;
 
 	memset(out, 0, sizeof(hostport_type));
+	memset(workbuf, 0, sizeof(workbuf));
 	/* Work on a copy of the input string. */
-	strncpy(workbuf, in, sizeof(workbuf));
+	strncpy(workbuf, in, sizeof(workbuf) - 1);
 	c = workbuf;
 	if (*c == '[') {
 		/* IPv6 addresses are enclosed in square brackets. */
@@ -374,15 +388,17 @@ int parse_hostport(
 			ret = getaddrinfo(srvname, NULL, &hints, &res0);
 			if (ret == 0) {
 				for (res = res0; res; res = res->ai_next) {
-					if (res->ai_family == AF_INET ||
-					    res->ai_family == AF_INET6) {
+					switch (res->ai_family) {
+					case AF_INET:
+					case AF_INET6:
 						/* Found a valid IPv4 or IPv6 address. */
 						memcpy(&out->IPaddress,
 						       res->ai_addr,
 						       res->ai_addrlen);
-						break;
+						goto found;
 					}
 				}
+found:
 				freeaddrinfo(res0);
 				if (res == NULL)
 					/* Didn't find an AF_INET or AF_INET6 address. */
@@ -404,24 +420,27 @@ int parse_hostport(
 			return UPNP_E_INVALID_URL;
 	} else
 		/* Port was not specified, use default port. */
-		port = 80;
+		port = 80u;
 	/* The length of the host and port string can be calculated by */
 	/* subtracting pointers. */
-	hostport_size = (size_t)(c - workbuf);
+	hostport_size = (size_t)c - (size_t)workbuf;
 	/* Fill in the 'out' information. */
-	if (af == AF_INET) {
-		sai4->sin_family = AF_INET;
+	switch (af) {
+	case AF_INET:
+		sai4->sin_family = (sa_family_t)af;
 		sai4->sin_port = htons(port);
 		ret = inet_pton(AF_INET, srvname, &sai4->sin_addr);
-	} else if (af == AF_INET6) {
-		sai6->sin6_family = AF_INET6;
+		break;
+	case AF_INET6:
+		sai6->sin6_family = (sa_family_t)af;
 		sai6->sin6_port = htons(port);
 		sai6->sin6_scope_id = gIF_INDEX;
 		ret = inet_pton(AF_INET6, srvname, &sai6->sin6_addr);
-	} else {
+		break;
+	default:
 		/* IP address was set by the hostname (getaddrinfo). */
 		/* Override port: */
-		if (out->IPaddress.ss_family == AF_INET)
+		if (out->IPaddress.ss_family == (sa_family_t)AF_INET)
 			sai4->sin_port = htons(port);
 		else
 			sai6->sin6_port = htons(port);
@@ -434,7 +453,6 @@ int parse_hostport(
 	out->text.buff = in;
 
 	return (int)hostport_size;
-	max = max;
 }
 
 /*!
@@ -455,19 +473,19 @@ static size_t parse_scheme(
 	/*! [out] Output parameter whose buffer is filled in with the scheme. */
 	token *out)
 {
-    size_t i = 0;
+    size_t i = (size_t)0;
 
-    out->size = 0;
+    out->size = (size_t)0;
     out->buff = NULL;
 
-    if( ( max == 0 ) || ( !isalpha( in[0] ) ) )
-        return 0;
+    if( ( max == (size_t)0 ) || ( !isalpha( in[0] ) ) )
+        return (size_t)0;
 
     i++;
     while( ( i < max ) && ( in[i] != ':' ) ) {
         if( !( isalnum( in[i] ) || ( in[i] == '+' ) || ( in[i] == '-' )
                || ( in[i] == '.' ) ) )
-            return 0;
+            return (size_t)0;
         i++;
     }
     if( i < max ) {
@@ -476,15 +494,15 @@ static size_t parse_scheme(
         return i;
     }
 
-    return 0;
+    return (size_t)0;
 }
 
 
 int remove_escaped_chars(INOUT char *in, INOUT size_t *size)
 {
-	size_t i = 0;
+	size_t i = (size_t)0;
 
-	for (i = 0; i < *size; i++) {
+	for (i = (size_t)0; i < *size; i++) {
 		replace_escaped(in, i, size);
 	}
 
@@ -492,68 +510,81 @@ int remove_escaped_chars(INOUT char *in, INOUT size_t *size)
 }
 
 
-int remove_dots(char *in, size_t size)
+static UPNP_INLINE int is_end_path(char c) {
+    switch (c) {
+	case '?':
+	case '#':
+	case '\0':
+	    return 1;
+    }
+    return 0;
+}
+
+
+/* This function directly implements the "Remove Dot Segments"
+ * algorithm described in RFC 3986 section 5.2.4. */
+int remove_dots(char *buf, size_t size)
 {
-    char *copyTo = in;
-    char *copyFrom = in;
-    char *max = in + size;
-    char **Segments = NULL;
-    int lastSegment = -1;
+    char *in = buf;
+    char *out = buf;
+    char *max = buf + size;
 
-    Segments = malloc( sizeof( char * ) * size );
+    while (!is_end_path(in[0])) {
+	assert (buf <= out);
+	assert (out <= in);
+	assert (in < max);
 
-    if( Segments == NULL )
-        return UPNP_E_OUTOF_MEMORY;
-
-    Segments[0] = NULL;
-    UpnpPrintf( UPNP_ALL, API, __FILE__, __LINE__,
-        "REMOVE_DOTS: before: %s\n", in );
-    while( ( copyFrom < max ) && ( *copyFrom != '?' )
-           && ( *copyFrom != '#' ) ) {
-
-        if( ( ( *copyFrom ) == '.' )
-            && ( ( copyFrom == in ) || ( *( copyFrom - 1 ) == '/' ) ) ) {
-            if( ( copyFrom + 1 == max )
-                || ( *( copyFrom + 1 ) == '/' ) ) {
-
-                copyFrom += 2;
-                continue;
-            } else if( ( *( copyFrom + 1 ) == '.' )
-                       && ( ( copyFrom + 2 == max )
-                            || ( *( copyFrom + 2 ) == '/' ) ) ) {
-                copyFrom += 3;
-
-                if( lastSegment > 0 ) {
-                    copyTo = Segments[--lastSegment];
-                } else {
-                    free( Segments );
-                    /*TRACE("ERROR RESOLVING URL, ../ at ROOT"); */
-                    return UPNP_E_INVALID_URL;
-                }
-                continue;
+        /* case 2.A: */
+        if (strncmp(in, "./", 2) == 0) {
+            in += 2;
+        } else if (strncmp(in, "../", 3) == 0) {
+            in += 3;
+        /* case 2.B: */
+        } else if (strncmp(in, "/./", 3) == 0) {
+            in += 2;
+        } else if (strncmp(in, "/.", 2) == 0 && is_end_path(in[2])) {
+            in += 1;
+	    in[0] = '/';
+        /* case 2.C: */
+        } else if (strncmp(in, "/../", 4) == 0 || (strncmp(in, "/..", 3) == 0 && is_end_path(in[3]))) {
+            /* Make the next character in the input buffer a '/': */
+            if (is_end_path(in[3])) { /* terminating "/.." case */
+                in += 2;
+                in[0] = '/';
+            } else { /* "/../" prefix case */
+                in += 3;
             }
-        }
-
-        if( ( *copyFrom ) == '/' ) {
-
-            lastSegment++;
-            Segments[lastSegment] = copyTo + 1;
-        }
-        ( *copyTo ) = ( *copyFrom );
-        copyTo++;
-        copyFrom++;
-    }
-    if( copyFrom < max ) {
-        while( copyFrom < max ) {
-            ( *copyTo ) = ( *copyFrom );
-            copyTo++;
-            copyFrom++;
+            /* Trim the last component from the output buffer, or empty it. */
+            while (buf < out)
+		if (*--out == '/')
+		    break;
+#ifdef DEBUG
+	    if (out < in)
+		out[0] = '\0';
+#endif
+        /* case 2.D: */
+        } else if (strncmp(in, ".", 1) == 0 && is_end_path(in[1])) {
+            in += 1;
+	} else if (strncmp(in, "..", 2) == 0 && is_end_path(in[2])) {
+            in += 2;
+        /* case 2.E */
+        } else {
+            /* move initial '/' character (if any) */
+            if (in[0] == '/')
+		*out++ = *in++;
+	    /* move first segment up to, but not including, the next '/' character */
+	    while (in < max && in[0] != '/' && !is_end_path(in[0]))
+		*out++ = *in++;
+#ifdef DEBUG
+	    if (out < in)
+		out[0] = '\0';
+#endif
         }
     }
-    ( *copyTo ) = 0;
-    free( Segments );
-    UpnpPrintf( UPNP_ALL, API, __FILE__, __LINE__,
-        "REMOVE_DOTS: after: %s\n", in );
+    while (in < max)
+	*out++ = *in++;
+    if (out < max)
+	out[0] = '\0';
     return UPNP_E_SUCCESS;
 }
 
@@ -562,117 +593,121 @@ char *resolve_rel_url(char *base_url, char *rel_url)
 {
     uri_type base;
     uri_type rel;
-    char temp_path = '/';
+    int rv;
 
-    size_t i = 0;
-    char *finger = NULL;
-
-    char *last_slash = NULL;
-
-    char *out = NULL;
-    char *out_finger = NULL;
-
-    if( base_url && rel_url ) {
-        out =
-            ( char * )malloc( strlen( base_url ) + strlen( rel_url ) + 2 );
-        out_finger = out;
-    } else {
-        if( rel_url )
-            return strdup( rel_url );
-        else
+    if (!base_url) {
+        if (!rel_url)
             return NULL;
+        return strdup(rel_url);
     }
 
-    if( out == NULL ) {
+    size_t len_rel = strlen(rel_url);
+    if (parse_uri(rel_url, len_rel, &rel) != HTTP_SUCCESS)
         return NULL;
+    if (rel.type == (enum uriType)ABSOLUTE)
+        return strdup(rel_url);
+
+    size_t len_base = strlen(base_url);
+    if ((parse_uri(base_url, len_base, &base) != HTTP_SUCCESS)
+            || (base.type != (enum uriType)ABSOLUTE))
+        return NULL;
+    if (len_rel == (size_t)0)
+        return strdup(base_url);
+
+    size_t len = len_base + len_rel + (size_t)2;
+    char *out = (char *)malloc(len);
+    if (out == NULL)
+        return NULL;
+    memset(out, 0, len);
+    char *out_finger = out;
+
+    /* scheme */
+    rv = snprintf(out_finger, len, "%.*s:", (int)base.scheme.size, base.scheme.buff);
+    if (rv < 0 || rv >= (int)len)
+        goto error;
+    out_finger += rv;
+    len -= (size_t)rv;
+
+    /* authority */
+    if (rel.hostport.text.size > (size_t)0) {
+        rv = snprintf(out_finger, len, "%s", rel_url);
+        if (rv < 0 || rv >= (int)len)
+            goto error;
+        return out;
+    }
+    if (base.hostport.text.size > (size_t)0) {
+	rv = snprintf(out_finger, len, "//%.*s", (int)base.hostport.text.size, base.hostport.text.buff);
+	if (rv < 0 || rv >= (int)len)
+	    goto error;
+	out_finger += rv;
+	len -= (size_t)rv;
     }
 
-    if( ( parse_uri( rel_url, strlen( rel_url ), &rel ) ) == HTTP_SUCCESS ) {
-
-        if( rel.type == ABSOLUTE ) {
-
-            strcpy( out, rel_url );
-        } else {
-
-            if( ( parse_uri( base_url, strlen( base_url ), &base ) ==
-                  HTTP_SUCCESS )
-                && ( base.type == ABSOLUTE ) ) {
-
-                if( strlen( rel_url ) == 0 ) {
-                    strcpy( out, base_url );
-                } else {
-                    memcpy( out, base.scheme.buff, base.scheme.size );
-                    out_finger += base.scheme.size;
-                    ( *out_finger ) = ':';
-                    out_finger++;
-
-                    if( rel.hostport.text.size > 0 ) {
-                        sprintf( out_finger, "%s", rel_url );
-                    } else {
-                        if( base.hostport.text.size > 0 ) {
-                            memcpy( out_finger, "//", 2 );
-                            out_finger += 2;
-                            memcpy( out_finger, base.hostport.text.buff,
-                                    base.hostport.text.size );
-                            out_finger += base.hostport.text.size;
-                        }
-
-                        if( rel.path_type == ABS_PATH ) {
-                            strcpy( out_finger, rel_url );
-
-                        } else {
-
-                            if( base.pathquery.size == 0 ) {
-                                base.pathquery.size = 1;
-                                base.pathquery.buff = &temp_path;
-                            }
-
-                            finger = out_finger;
-                            last_slash = finger;
-                            i = 0;
-                            while( ( i < base.pathquery.size ) &&
-                                   ( base.pathquery.buff[i] != '?' ) ) {
-                                ( *finger ) = base.pathquery.buff[i];
-                                if( base.pathquery.buff[i] == '/' )
-                                    last_slash = finger + 1;
-                                i++;
-                                finger++;
-
-                            }
-                            strcpy( last_slash, rel_url );
-                            if( remove_dots( out_finger,
-                                             strlen( out_finger ) ) !=
-                                UPNP_E_SUCCESS ) {
-                                free(out);
-                                /* free(rel_url); */
-                                return NULL;
-                            }
-                        }
-
-                    }
-                }
-            } else {
-                free(out);
-                /* free(rel_url); */
-                return NULL;
-            }
-        }
+    /* path */
+    char *path = out_finger;
+    if (rel.path_type == (enum pathType)ABS_PATH) {
+	rv = snprintf(out_finger, len, "%s", rel_url);
+    } else if (base.pathquery.size == (size_t)0) {
+	rv = snprintf(out_finger, len, "/%s", rel_url);
     } else {
-        free(out);
-        /* free(rel_url); */          
-        return NULL;
+	if (rel.pathquery.size == (size_t)0) {
+	    rv = snprintf(out_finger, len, "%.*s", (int)base.pathquery.size, base.pathquery.buff);
+	} else {
+	    if (len < base.pathquery.size)
+		goto error;
+	    size_t i = (size_t)0, prefix = (size_t)1;
+	    while (i < base.pathquery.size) {
+		out_finger[i] = base.pathquery.buff[i];
+		switch (base.pathquery.buff[i++]) {
+		    case '/':
+			prefix = i;
+			/* fall-through */
+		    default:
+			continue;
+		    case '?': /* query */
+			if (rel.pathquery.buff[0] == '?')
+			    prefix = --i;
+		}
+		break;
+	    }
+	    out_finger += prefix;
+	    len -= prefix;
+	    rv = snprintf(out_finger, len, "%.*s", (int)rel.pathquery.size, rel.pathquery.buff);
+	}
+	if (rv < 0 || rv >= (int)len)
+	    goto error;
+	out_finger += rv;
+	len -= (size_t)rv;
+
+	/* fragment */
+	if (rel.fragment.size > (size_t)0)
+	    rv = snprintf(out_finger, len, "#%.*s", (int)rel.fragment.size, rel.fragment.buff);
+	else if (base.fragment.size > (size_t)0)
+	    rv = snprintf(out_finger, len, "#%.*s", (int)base.fragment.size, base.fragment.buff);
+	else
+	    rv = 0;
     }
+    if (rv < 0 || rv >= (int)len)
+	goto error;
+    out_finger += rv;
+    len -= (size_t)rv;
+
+    if (remove_dots(path, (size_t)(out_finger - path)) != UPNP_E_SUCCESS)
+	goto error;
 
-    /* free(rel_url); */
     return out;
+
+error:
+    free(out);
+    return NULL;
 }
 
 
 int parse_uri(const char *in, size_t max, uri_type *out)
 {
 	int begin_path = 0;
-	size_t begin_hostport = 0;
-	size_t begin_fragment = 0;
+	size_t begin_hostport = (size_t)0;
+	size_t begin_fragment = (size_t)0;
 
 	begin_hostport = parse_scheme(in, max, &out->scheme);
 	if (begin_hostport) {
@@ -683,12 +718,11 @@ int parse_uri(const char *in, size_t max, uri_type *out)
 		out->type = RELATIVE;
 		out->path_type = REL_PATH;
 	}
-	if (begin_hostport + 1 < max &&
+	if (begin_hostport + (size_t)1 < max &&
 	    in[begin_hostport] == '/' &&
-	    in[begin_hostport + 1] == '/') {
-		begin_hostport += 2;
+	    in[begin_hostport + (size_t)1] == '/') {
+		begin_hostport += (size_t)2;
 		begin_path = parse_hostport(&in[begin_hostport],
-			max - begin_hostport,
 			&out->hostport);
 		if (begin_path >= 0) {
 			begin_path += (int)begin_hostport;
@@ -710,7 +744,7 @@ int parse_uri(const char *in, size_t max, uri_type *out)
 			   &out->fragment);
 	} else {
 		out->fragment.buff = NULL;
-		out->fragment.size = 0;
+		out->fragment.size = (size_t)0;
 	}
 
 	return HTTP_SUCCESS;
@@ -724,10 +758,10 @@ int parse_uri_and_unescape(char *in, size_t max, uri_type *out)
 		return ret;
 	}
 
-	if (out->pathquery.size > 0) {
+	if (out->pathquery.size > (size_t)0) {
 		remove_escaped_chars((char *)out->pathquery.buff, &out->pathquery.size);
 	}
-	if (out->fragment.size > 0) {
+	if (out->fragment.size > (size_t)0) {
 		remove_escaped_chars((char *)out->fragment.buff, &out->fragment.size);
 	}
 
