@@ -467,16 +467,16 @@ int unique_service_name(char *cmd, SsdpEvent *Evt)
 		else
 			return -1;
 		if (ptr3 != NULL) {
-			if (strlen("uuid:") + strlen(ptr3 + 1) >= sizeof(Evt->UDN))
+			if (strlen("uuid:") + strlen(ptr3 + 1) >= sizeof Evt->UDN)
 				return -1;
-			snprintf(Evt->UDN, sizeof(Evt->UDN), "uuid:%s",
-				ptr3 + 1);
+			snprintf(Evt->UDN, sizeof Evt->UDN, "uuid:%s", ptr3 + 1);
 		}
 		else
 			return -1;
 		ptr1 = strstr(cmd, ":");
 		if (ptr1 != NULL) {
 			n = (size_t)ptr3 - (size_t)ptr1;
+			n = n >= sizeof TempBuf ? sizeof TempBuf - 1 : n;
 			strncpy(TempBuf, ptr1, n);
 			TempBuf[n] = '\0';
 			if (strlen("urn") + strlen(TempBuf) >= sizeof(Evt->DeviceType))
@@ -490,27 +490,28 @@ int unique_service_name(char *cmd, SsdpEvent *Evt)
 	if ((TempPtr = strstr(cmd, "uuid")) != NULL) {
 		if ((Ptr = strstr(cmd, "::")) != NULL) {
 			n = (size_t)Ptr - (size_t)TempPtr;
+			n = n >= sizeof Evt->UDN ? sizeof Evt->UDN - 1 : n;
 			strncpy(Evt->UDN, TempPtr, n);
 			Evt->UDN[n] = '\0';
 		} else {
 			memset(Evt->UDN, 0, sizeof(Evt->UDN));
-			strncpy(Evt->UDN, TempPtr, sizeof(Evt->UDN) - 1);
+			strncpy(Evt->UDN, TempPtr, sizeof Evt->UDN - 1);
 		}
 		CommandFound = 1;
 	}
 	if (strstr(cmd, "urn:") != NULL && strstr(cmd, ":service:") != NULL) {
 		if ((TempPtr = strstr(cmd, "urn")) != NULL) {
-			memset(Evt->ServiceType, 0, sizeof(Evt->ServiceType));
+			memset(Evt->ServiceType, 0, sizeof Evt->ServiceType);
 			strncpy(Evt->ServiceType, TempPtr,
-				sizeof(Evt->ServiceType) - 1);
+				sizeof Evt->ServiceType - 1);
 			CommandFound = 1;
 		}
 	}
 	if (strstr(cmd, "urn:") != NULL && strstr(cmd, ":device:") != NULL) {
 		if ((TempPtr = strstr(cmd, "urn")) != NULL) {
-			memset(Evt->DeviceType, 0, sizeof(Evt->DeviceType));
+			memset(Evt->DeviceType, 0, sizeof Evt->DeviceType);
 			strncpy(Evt->DeviceType, TempPtr,
-				sizeof(Evt->DeviceType) - 1);
+				sizeof Evt->DeviceType - 1);
 			CommandFound = 1;
 		}
 	}
@@ -518,6 +519,7 @@ int unique_service_name(char *cmd, SsdpEvent *Evt)
 		/* Everything before "::upnp::rootdevice" is the UDN. */
 		if (TempPtr != cmd) {
 			n = (size_t)TempPtr - (size_t)cmd;
+			n = n >= sizeof Evt->UDN ? sizeof Evt->UDN - 1 : n;
 			strncpy(Evt->UDN, cmd, n);
 			Evt->UDN[n] = 0;
 			CommandFound = 1;
@@ -812,7 +814,7 @@ static int create_ssdp_sock_v4(
 		ret = UPNP_E_SOCKET_ERROR;
 		goto error_handler;
 	}
-#if defined(BSD) || defined(__OSX__) || defined(__APPLE__)
+#if (defined(BSD) && !defined(__GNU__)) || defined(__OSX__) || defined(__APPLE__)
 	onOff = 1;
 	ret = setsockopt(*ssdpSock, SOL_SOCKET, SO_REUSEPORT,
 			 (char *)&onOff, sizeof(onOff));
@@ -864,8 +866,8 @@ static int create_ssdp_sock_v4(
 		/* This is probably not a critical error, so let's continue. */
 	}
 	/* result is not checked becuase it will fail in WinMe and Win9x. */
-	ret = setsockopt(*ssdpSock, IPPROTO_IP,
-			 IP_MULTICAST_TTL, &ttl, sizeof(ttl));
+	setsockopt(*ssdpSock, IPPROTO_IP,
+		IP_MULTICAST_TTL, &ttl, sizeof(ttl));
 	onOff = 1;
 	ret = setsockopt(*ssdpSock, SOL_SOCKET, SO_BROADCAST,
 			 (char *)&onOff, sizeof(onOff));
@@ -881,11 +883,6 @@ static int create_ssdp_sock_v4(
 
 error_handler:
 	if (ret != UPNP_E_SUCCESS) {
-		if (shutdown(*ssdpSock, SD_BOTH) == -1) {
-			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, SSDP, __FILE__, __LINE__,
-				   "Error in shutdown: %s\n", errorBuffer);
-                }
                 UpnpCloseSocket(*ssdpSock);
 	}
 
@@ -954,7 +951,7 @@ static int create_ssdp_sock_v6(
 		ret = UPNP_E_SOCKET_ERROR;
 		goto error_handler;
 	}
-#if defined(BSD) || defined(__OSX__) || defined(__APPLE__)
+#if (defined(BSD) && !defined(__GNU__)) || defined(__OSX__) || defined(__APPLE__)
 	onOff = 1;
 	ret = setsockopt(*ssdpSock, SOL_SOCKET, SO_REUSEPORT,
 			 (char *)&onOff, sizeof(onOff));
@@ -1021,11 +1018,6 @@ static int create_ssdp_sock_v6(
 
 error_handler:
 	if (ret != UPNP_E_SUCCESS) {
-		if (shutdown(*ssdpSock, SD_BOTH) == -1) {
-			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, SSDP, __FILE__, __LINE__,
-				   "Error in shutdown: %s\n", errorBuffer);
-		}
 		UpnpCloseSocket(*ssdpSock);
 	}
 
@@ -1067,7 +1059,7 @@ static int create_ssdp_sock_v6_ula_gua(
 		ret = UPNP_E_SOCKET_ERROR;
 		goto error_handler;
 	}
-#if defined(BSD) || defined(__OSX__) || defined(__APPLE__)
+#if (defined(BSD) && !defined(__GNU__)) || defined(__OSX__) || defined(__APPLE__)
 	onOff = 1;
 	ret = setsockopt(*ssdpSock, SOL_SOCKET, SO_REUSEPORT,
 			 (char *)&onOff, sizeof(onOff));
@@ -1135,11 +1127,6 @@ static int create_ssdp_sock_v6_ula_gua(
 
 error_handler:
 	if (ret != UPNP_E_SUCCESS) {
-		if (shutdown(*ssdpSock, SD_BOTH) == -1) {
-			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, SSDP, __FILE__, __LINE__,
-				   "Error in shutdown: %s\n", errorBuffer);
-		}
 		UpnpCloseSocket(*ssdpSock);
 	}
 
@@ -1199,7 +1186,6 @@ int get_ssdp_sockets(MiniServerSockArray * out)
 	if (strlen(gIF_IPV6) > (size_t)0) {
 		retVal = create_ssdp_sock_reqv6(&out->ssdpReqSock6);
 		if (retVal != UPNP_E_SUCCESS) {
-			shutdown(out->ssdpReqSock4, SD_BOTH);
 			UpnpCloseSocket(out->ssdpReqSock4);
 			return retVal;
 		}
@@ -1214,9 +1200,7 @@ int get_ssdp_sockets(MiniServerSockArray * out)
 		retVal = create_ssdp_sock_v4(&out->ssdpSock4);
 		if (retVal != UPNP_E_SUCCESS) {
 #ifdef INCLUDE_CLIENT_APIS
-			shutdown(out->ssdpReqSock4, SD_BOTH);
 			UpnpCloseSocket(out->ssdpReqSock4);
-			shutdown(out->ssdpReqSock6, SD_BOTH);
 			UpnpCloseSocket(out->ssdpReqSock6);
 #endif /* INCLUDE_CLIENT_APIS */
 			return retVal;
@@ -1228,12 +1212,9 @@ int get_ssdp_sockets(MiniServerSockArray * out)
 	if (strlen(gIF_IPV6) > (size_t)0) {
 		retVal = create_ssdp_sock_v6(&out->ssdpSock6);
 		if (retVal != UPNP_E_SUCCESS) {
-			shutdown(out->ssdpSock4, SD_BOTH);
 			UpnpCloseSocket(out->ssdpSock4);
 #ifdef INCLUDE_CLIENT_APIS
-			shutdown(out->ssdpReqSock4, SD_BOTH);
 			UpnpCloseSocket(out->ssdpReqSock4);
-			shutdown(out->ssdpReqSock6, SD_BOTH);
 			UpnpCloseSocket(out->ssdpReqSock6);
 #endif /* INCLUDE_CLIENT_APIS */
 			return retVal;
@@ -1243,14 +1224,10 @@ int get_ssdp_sockets(MiniServerSockArray * out)
 	if (strlen(gIF_IPV6_ULA_GUA) > (size_t)0) {
 		retVal = create_ssdp_sock_v6_ula_gua(&out->ssdpSock6UlaGua);
 		if (retVal != UPNP_E_SUCCESS) {
-			shutdown(out->ssdpSock4, SD_BOTH);
 			UpnpCloseSocket(out->ssdpSock4);
-			shutdown(out->ssdpSock6, SD_BOTH);
 			UpnpCloseSocket(out->ssdpSock6);
 #ifdef INCLUDE_CLIENT_APIS
-			shutdown(out->ssdpReqSock4, SD_BOTH);
 			UpnpCloseSocket(out->ssdpReqSock4);
-			shutdown(out->ssdpReqSock6, SD_BOTH);
 			UpnpCloseSocket(out->ssdpReqSock6);
 #endif /* INCLUDE_CLIENT_APIS */
 			return retVal;
